Require Import Classical.
Require Import semantics.
Require Import Coq.Program.Equality.

(** The language **)

Inductive div: Set :=
| Div: div
| Conv: div.

(** Big-step semantics **)

Inductive evald: term -> div -> term -> div -> Prop :=
| evald_const: forall c,
    evald (Const c) Conv (Const c) Conv
| evald_fun: forall x a,
    evald (Fun x a) Conv (Fun x a) Conv
| evald_app: forall a b x c vb v d d' d'',
    evald a Conv (Fun x c) d ->
    evald b d vb d' ->
    evald (subst x vb c) d' v d'' ->
    evald (App a b) Conv v d''
| evald_div: forall a b,
    evald a Div b Div.

Lemma evald_conv :
  forall a b d,
    evald a Conv b d ->
    d = Conv.
Proof.
  induction 1; auto. subst. auto.
Qed.

Hint Constructors eval evald.

Lemma evald_eval :
  forall a b,
    eval a b <->
    evald a Conv b Conv.
Proof.
  intros. split; intro E; dependent induction E; eauto.
  apply evald_conv in E1. subst.
  apply evald_conv in E2. subst.
  econstructor; auto.
Qed.

CoInductive coevald: term -> div -> term -> div -> Prop :=
| coevald_const: forall c,
    coevald (Const c) Conv (Const c) Conv
| coevald_fun: forall x a,
    coevald (Fun x a) Conv (Fun x a) Conv
| coevald_app: forall a b x c vb v d d' d'',
    coevald a Conv (Fun x c) d ->
    coevald b d vb d' ->
    coevald (subst x vb c) d' v d'' ->
    coevald (App a b) Conv v d''
| coevald_div: forall a b,
    coevald a Div b Div.

Lemma evald_coevald :
  forall a d b d',
    evald a d b d' -> coevald a d b d'.
Proof.
  induction 1; econstructor; eauto.
Qed.

(** Verbatim reuse modulo constructor names. *)
Lemma coevald_omega :
  forall b,
    coevald omega Conv b Div.
Proof.
  cofix COINDHYP. intro; unfold omega. eapply coevald_app.
  unfold delta. apply coevald_fun.
  unfold delta. apply coevald_fun.
  simpl. fold delta. fold omega. apply COINDHYP.
Qed.

Hint Constructors evalinf.

Lemma coevald_notevald_coevald_div :
  forall a b,
    coevald a Conv b Conv ->
    ~ (evald a Conv b Conv) ->
    coevald a Conv b Div.
Proof.
  cofix COINDHYP. intros.
  inversion H; subst.
  - destruct H0; auto.
  - destruct H0; auto.
  - destruct d'. inversion H3.
    destruct d. inversion H2.
    elim (classic (evald a0 Conv (Fun x c) Conv)); intro.
    elim (classic (evald b0 Conv vb Conv)); intro.
    elim (classic (evald (subst x vb c) Conv b Conv)); intro.
    econstructor; eauto.
    econstructor; eauto.
    econstructor. eassumption. apply COINDHYP. eauto. eauto. econstructor.
    econstructor. apply COINDHYP; eauto. econstructor. econstructor.
  Grab Existential Variables. constructor. constructor.
Qed.

Lemma coevald_evald_or_evald_div :
  forall a v, coevald a Conv v Conv -> evald a Conv v Conv \/ coevald a Conv v Div.
Proof.
  intros. elim (classic (evald a Conv v Conv)); intro.
  left; auto.
  right. eapply coevald_notevald_coevald_div; eauto.
Qed.

Lemma coevald_div_evalinf :
  forall a b,
    coevald a Conv b Div <-> evalinf a.
Proof.
  intros. split.
  {
    generalize dependent b. generalize dependent a.
    cofix COINDHYP. intros.
    inversion H; subst.
    destruct d. econstructor. apply COINDHYP in H0; eauto.
    destruct d'.
      apply coevald_evald_or_evald_div in H0.
      destruct H0. econstructor 2. apply evald_eval. apply H0.
      apply COINDHYP in H1. apply H1.
      econstructor. apply COINDHYP in H0. assumption.
    apply coevald_evald_or_evald_div in H0.
    apply coevald_evald_or_evald_div in H1.
    destruct H0. destruct H1. econstructor 3.
      apply evald_eval; eassumption.
      apply evald_eval; eassumption.
      apply COINDHYP in H2. assumption.
      apply COINDHYP in H1.
        econstructor 2. apply evald_eval; eassumption. auto.
      apply COINDHYP in H0. econstructor; assumption.
  }{
    generalize dependent b. generalize dependent a.
    cofix COINDHYP. intros.
    inversion H; subst.
    - econstructor. apply COINDHYP. apply H0. econstructor. econstructor.
    - apply evald_eval in H0. apply evald_coevald in H0.
      econstructor; eauto. econstructor.
    - econstructor.
      apply evald_coevald. apply evald_eval. apply H0.
      apply evald_coevald. apply evald_eval. apply H1.
      eauto.
  }
  Grab Existential Variables.
  constructor. constructor. constructor. constructor. constructor. constructor.
  constructor.
Qed.
