Require Import Classical.
Require Import semantics.
Require Import typing.
Require Import div.
Require Import Coq.Program.Equality.

(** The type system **)

Hint Constructors has_type coevald isvalue.

Lemma evald_preservation:
  forall a v, evald a Conv v Conv ->
  forall t, has_type env_empty a t -> has_type env_empty v t.
Proof.
  intros ? ? E. dependent induction E; intros.
  auto.
  auto.
  inversion H. subst.
  assert (E1':=E1). assert (E2':=E2).
  apply evald_conv in E1'. subst. apply evald_conv in E2'. subst.
  assert (has_type env_empty (Fun x c) (Tarrow t1 t)).
  eauto.
  apply IHE3; auto.
  apply has_type_subst with t1.
  inversion H0. subst. auto.
  auto.
Qed.

Lemma evald_or_not_evald:
  forall a,
  (forall v, ~evald a Conv v Conv) \/ (exists v, evald a Conv v Conv).
Proof.
  intro. elim (classic (forall v, ~evald a Conv v Conv)); intro.
  left; auto.
  right. apply not_all_not_ex. assumption.
Qed.

Lemma coevald_progress :
  forall a t,
    has_type env_empty a t -> (forall v, ~evald a Conv v Conv) ->
    forall v, coevald a Conv v Div.
Proof.
  cofix CIH. intros. inversion H; subst.
  - rewrite env_get_empty in H1. discriminate.
  - specialize (H0 (Const n)). destruct H0; auto.
  - specialize (H0 (Fun x a0)). destruct H0; auto.
  - elim (evald_or_not_evald a0); intro.
    { eapply CIH in H1; auto. econstructor. apply H1.
      constructor. constructor.
    }
    destruct H3.
    assert (has_type env_empty x (Tarrow t1 t)).
    eapply evald_preservation; eauto.
    assert (exists x' c, x = Fun x' c).
    { assert (isvalue x).
      apply evald_eval in H3. apply eval_isvalue in H3. auto.
      inversion H5; subst. inversion H4. exists x0. exists a. auto.
    }
    do 2 destruct H5. subst.
    elim (evald_or_not_evald b); intro.
    { eapply CIH in H5. Focus 2. apply H2. econstructor.
      apply evald_coevald. apply H3. eassumption. constructor.
    }
    destruct H5.
    assert (has_type env_empty x t1).
    { eapply evald_preservation; eauto. }
    inversion H4; subst.
    eapply has_type_subst in H6. Focus 2. apply H9.
    elim (evald_or_not_evald (subst x0 x x1)); intro.
    { eapply CIH in H7. Focus 2. apply H6.
      econstructor.
      apply evald_coevald. eassumption.
      apply evald_coevald. eassumption.
      eassumption.
    }
    destruct H7. specialize (H0 x2). destruct H0.
    econstructor; eassumption.
    Grab Existential Variables.
    constructor. constructor. constructor. constructor.
    constructor. constructor. constructor.
Qed.

Lemma coevald_sound :
  forall a t,
    has_type env_empty a t -> exists v d, coevald a Conv v d.
Proof.
  intros. elim (evald_or_not_evald a); intro.
  eexists (Var 0); exists Div. eapply coevald_progress; eauto.
  destruct H0. apply evald_coevald in H0. eauto.
Qed.

Hint Constructors eval evalinf.
Hint Unfold subst.

Ltac Appl := eapply evalinf_app_l.
Ltac Appr := eapply evalinf_app_r.
Ltac Appf := eapply evalinf_app_f.

Ltac norm := repeat (econstructor; simpl; eauto; simpl); simpl.

Lemma evalinf_filinski :
  evalinf filinski.
Proof.
  unfold filinski. unfold Y. unfold filinski1.
  Appf. norm. norm. simpl.
  Appr; eauto.
  Appf; eauto. simpl.
  Appr; eauto.
  Appf; eauto. simpl.
  Appf; eauto. norm. simpl.
  cofix IH.
  Appr; eauto.
  Appf; eauto. simpl.
  Appf; eauto. norm. simpl.
  apply IH.
Qed.

Lemma coevald_filinski :
  forall v, coevald filinski Conv v Div.
Proof.
  intro. apply coevald_div_evalinf. apply evalinf_filinski.
Qed.
